"""Provide imports for confounds sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
# License: AGPL

from .fmriprep_confound_remover import fMRIPrepConfoundRemover
