.. _Fede Raimondo: https://fraimondo.github.io
.. _Kaustubh R. Patil: https://github.com/kaurao
.. _Leonard Sasse: https://github.com/LeSasse
.. _Amir Omidvarnia: https://github.com/omidvarnia
.. _Synchon Mandal: https://github.com/synchon
.. _Benjamin Poldrack: https://github.com/bpoldrack
.. _Vera Komeyer: https://github.com/verakye
.. _Xuan Li: https://github.com/xuanli-ac
.. _Sami Hamdan: https://github.com/samihamdan
.. _Nicolas Nieto: https://github.com/N-Nieto
.. _Felix Hoffstaedter: https://github.com/FeHoff
