.. This (-*- rst -*-) format file contains commonly used link targets
   and name substitutions.  It may be included in many files,
   therefore it should only contain link targets and name
   substitutions.  Try grepping for "^\.\. _" to find plausible
   candidates for this list.

.. NOTE: reST targets are
   __not_case_sensitive__, so only one target definition is needed for
   nipy, NIPY, Nipy, etc...


.. _`AML`: https://www.fz-juelich.de/inm/inm-7/EN/Forschung/Applied%20Machine%20Learning/_node.html
.. _`INM-7`: https://www.fz-juelich.de/inm/inm-7/EN/Home/home_node.html
.. _`julearn`: https://juaml.github.io/julearn

.. _`pandas`: https://pandas.pydata.org
.. _`pandas.DataFrame` : https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html

.. _`Python`: https://python.org
.. _`seaborn`: https://seaborn.pydata.org
.. _`matplotlib`: https://matplotlib.org
.. _`fMRIPrep`: https://fmriprep.org
.. _`nilearn`: https://nilearn.github.io
.. _`nipype`: https://nipype.readthedocs.io
.. _`datalad`: https://datalad.org

.. _`venv`: https://docs.python.org/3/tutorial/venv.html
.. _`conda env`: https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-environments.html

.. _`Github`: https://github.com/
.. _`junifer Github`: https://github.com/juaml/junifer
.. _`junifer Discussions`: https://github.com/juaml/junifer/discussions

.. _`Wikipedia`: https://wikipedia.org
.. _`YAML`: https://yaml.org

.. _`setuptools_scm`: https://github.com/pypa/setuptools_scm/

.. _`sphinx gallery`: https://sphinx-gallery.github.io/stable/index.html
.. _`sphinx reST reference`: https://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html#inline-markup

.. _`HTCondor`: https://research.cs.wisc.edu/htcondor/
.. _`SLURM`: https://slurm.schedmd.com
.. _`GNU Parallel`: https://www.gnu.org/software/parallel/