.table-wrapper.colwidths-auto table td, .table-wrapper.colwidths-auto table th {
  white-space: nowrap;
}
.table-wrapper.colwidths-auto table td a.reference.internal {
  overflow-wrap: normal;
}

/* Style the button that is used to open and close the collapsible content */
.version-select {
  cursor: pointer;
}

/* Style the collapsible content. Note: hidden by default */
#versions {
  display: none;
  overflow: hidden;
}